from marshmallow import ValidationError
from freezegun import freeze_time
from test.backendtestcase import TestCase
from sqlalchemy.orm import Session
from src.cs_models.database import engine, operations
from src.cs_models.resources.Merger.models import MergerModel
from src.cs_models.resources.Merger.schemas import MergerResourceSchema
from src.cs_models.resources.File.models import FileModel
from src.cs_models.resources.File.schemas import FileResourceSchema
from src.cs_models.resources.CompanySEC.models import CompanySECModel
from src.cs_models.resources.CompanySEC.schemas import CompanySECResourceSchema
from src.cs_models.resources.CompanyOUS.models import CompanyOUSModel
from src.cs_models.resources.CompanyOUS.schemas import CompanyOUSResourceSchema


@freeze_time("2020-01-01")
class MergerResourceTestCase(TestCase):
    def setUp(self):
        super(MergerResourceTestCase, self).setUp()
        with operations.session_scope() as session:

            self.company_sec1 = operations.create(
                session=session,
                model=CompanySECModel,
                schema=CompanySECResourceSchema,
                obj={
                    'cik_str': '1750',
                    'ticker': 'ABBV',
                    'title': 'Abbvie',
                }
            )

            self.company_sec2 = operations.create(
                session=session,
                model=CompanySECModel,
                schema=CompanySECResourceSchema,
                obj={
                    'cik_str': '1760',
                    'ticker': 'MYL',
                    'title': 'Mylan N.V.',
                }
            )

            self.file1 = operations.create(
                session=session,
                model=FileModel,
                schema=FileResourceSchema,
                obj={
                    'file_format': 'html',
                    's3_bucket_name': 'dma',
                    's3_key_name': '1/dma.html',
                },
            )

            self.merger1 = operations.create(
                session=session,
                model=MergerModel,
                schema=MergerResourceSchema,
                obj={
                    "target_sec_id": self.company_sec1.id,
                    "acquirer_sec_id": self.company_sec2.id,
                    "deal_value": 1500000000,
                    "type": "Cash",
                    "announcement_date": "02/15/2020",
                    "offer_price": 10.30,
                    "market_price": 8.90,
                    "dma_file_id": self.file1.id,
                },
            )

            self.valid_data = {
                        "target_sec_id": self.company_sec1.id,
                        "acquirer_sec_id": self.company_sec2.id,
                        "deal_value": 20000000,
                        "type": "Cash",
                        "announcement_date": "01/01/2020",
                        "offer_price": 15.75,
                        "market_price": 11.43,
                        "dma_file_id": self.file1.id,
                    }

    def test_create_inside_context_manager(self):
        with operations.session_scope() as session:

            self.file2 = operations.create(
                session=session,
                model=FileModel,
                schema=FileResourceSchema,
                obj={
                    'file_format': 'html',
                    's3_bucket_name': 'dma',
                    's3_key_name': '2/dma.html',
                },
            )

            self.company_sec3 = operations.create(
                session=session,
                model=CompanySECModel,
                schema=CompanySECResourceSchema,
                obj={
                    'cik_str': '1770',
                    'ticker': 'LLY',
                    'title': 'Eli Lily',
                }
            )

            self.company_sec4 = operations.create(
                session=session,
                model=CompanySECModel,
                schema=CompanySECResourceSchema,
                obj={
                    'cik_str': '1780',
                    'ticker': 'JNJ',
                    'title': 'Johnson & Johnson',
                }
            )

            instance = operations.create(
                session=session,
                model=MergerModel,
                schema=MergerResourceSchema,
                obj={
                    "target_sec_id": self.company_sec3.id,
                    "acquirer_sec_id": self.company_sec4.id,
                    "deal_value": 148575940,
                    "type": "Cash",
                    "announcement_date": "05/31/2020",
                    "offer_price": 155.50,
                    "market_price": 138.90,
                    "dma_file_id": self.file2.id,
                },
            )
            self.assertEqual(
                MergerResourceSchema().dump(instance),
                {
                    "id": 2,
                    "target_sec_id": 3,
                    "target_ous_id": None,
                    "acquirer_sec_id": 4,
                    "acquirer_ous_id": None,
                    "deal_value": 148575940,
                    "announcement_date": '2020-05-31T00:00:00+00:00',
                    "type": "Cash",
                    "offer_price": 155.50,
                    "market_price": 138.90,
                    "dma_file_id": 2,
                    "updated_at": '2020-01-01T00:00:00+00:00',
                },
            )

    def test_create(self):
        session = Session(bind=engine)
        try:
            self.file3 = operations.create(
                session=session,
                model=FileModel,
                schema=FileResourceSchema,
                obj={
                    'file_format': 'html',
                    's3_bucket_name': 'dma',
                    's3_key_name': '3/dma.html',
                },
            )

            self.company_sec5 = operations.create(
                session=session,
                model=CompanySECModel,
                schema=CompanySECResourceSchema,
                obj={
                    'cik_str': '1790',
                    'ticker': 'PTLA',
                    'title': 'Portola Pharma',
                }
            )

            self.company_sec6 = operations.create(
                session=session,
                model=CompanySECModel,
                schema=CompanySECResourceSchema,
                obj={
                    'cik_str': '1800',
                    'ticker': 'NVS',
                    'title': 'Novartis',
                }
            )
            instance = operations.create(
                session=session,
                model=MergerModel,
                schema=MergerResourceSchema,
                obj={
                    "target_sec_id": self.company_sec5.id,
                    "acquirer_sec_id": self.company_sec6.id,
                    "deal_value": 2000000,
                    "type": "Cash",
                    "announcement_date": "02/01/2020",
                    "offer_price": 120,
                    "market_price": 105,
                    "dma_file_id": self.file3.id,
                },
            )
            self.assertEqual(
                MergerResourceSchema().dump(instance),
                {
                    "id": 2,
                    "target_sec_id": 3,
                    "target_ous_id": None,
                    "acquirer_sec_id": 4,
                    "acquirer_ous_id": None,
                    "deal_value": 2000000,
                    "type": "Cash",
                    "announcement_date": "2020-02-01T00:00:00+00:00",
                    "offer_price": 120.0,
                    "market_price": 105.0,
                    "dma_file_id": 2,
                    "updated_at": '2020-01-01T00:00:00+00:00',
                },
            )

            # This commit is not really needed since
            # we commit in `create` but is done for consistency.
            session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

    def test_create_validation_error(self):
        session = Session(bind=engine)
        self.file3 = operations.create(
            session=session,
            model=FileModel,
            schema=FileResourceSchema,
            obj={
                'file_format': 'html',
                's3_bucket_name': 'dma',
                's3_key_name': '3/dma.html',
            },
        )

        self.company_sec5 = operations.create(
            session=session,
            model=CompanySECModel,
            schema=CompanySECResourceSchema,
            obj={
                'cik_str': '1790',
                'ticker': 'PTLA',
                'title': 'Portola Pharma',
            }
        )

        self.company_sec6 = operations.create(
            session=session,
            model=CompanySECModel,
            schema=CompanySECResourceSchema,
            obj={
                'cik_str': '1800',
                'ticker': 'NVS',
                'title': 'Novartis',
            }
        )
        self.assertRaises(
            ValidationError,
            operations.create,
            session,
            MergerModel,
            MergerResourceSchema,
            {
                "target_sec_id": self.company_sec5.id,
                "acquirer_sec_id": self.company_sec6.id,
                "deal_value": 2000000,
                "type": "Cash",
                "offer_price": 120,
                "market_price": 105,
                "dma_file_id": self.file3.id,
            },
        )

    def test_update_or_create(self):
        with operations.session_scope() as session:
            self.file3 = operations.create(
                session=session,
                model=FileModel,
                schema=FileResourceSchema,
                obj={
                    'file_format': 'html',
                    's3_bucket_name': 'dma',
                    's3_key_name': '3/dma.html',
                },
            )

            self.company_sec5 = operations.create(
                session=session,
                model=CompanySECModel,
                schema=CompanySECResourceSchema,
                obj={
                    'cik_str': '1790',
                    'ticker': 'PTLA',
                    'title': 'Portola Pharma',
                }
            )

            self.company_sec6 = operations.create(
                session=session,
                model=CompanySECModel,
                schema=CompanySECResourceSchema,
                obj={
                    'cik_str': '1800',
                    'ticker': 'NVS',
                    'title': 'Novartis',
                }
            )
            instance, created = operations.update_or_create(
                session=session,
                model=MergerModel,
                schema=MergerResourceSchema,
                query_fields=["target_sec_id", "target_ous_id",
                              "acquirer_sec_id", "acquirer_ous_id",
                              "announcement_date"],
                obj={
                    "target_sec_id": self.company_sec5.id,
                    "acquirer_sec_id": self.company_sec6.id,
                    "deal_value": 2000000,
                    "type": "Cash",
                    "announcement_date": "02/14/2020",
                    "offer_price": 120,
                    "market_price": 105,
                    "dma_file_id": self.file3.id,
                },
            )
            self.assertEqual(
                MergerResourceSchema().dump(instance),
                {
                    "id": 2,
                    "target_sec_id": 3,
                    "target_ous_id": None,
                    "acquirer_sec_id": 4,
                    "acquirer_ous_id": None,
                    "deal_value": 2000000,
                    "type": "Cash",
                    "announcement_date": "2020-02-14T00:00:00+00:00",
                    "offer_price": 120.0,
                    "market_price": 105.0,
                    "dma_file_id": 2,
                    "updated_at": '2020-01-01T00:00:00+00:00',
                },
            )
            self.assertTrue(created)

            instance, created = operations.update_or_create(
                session=session,
                model=MergerModel,
                schema=MergerResourceSchema,
                query_fields=["target_sec_id", "target_ous_id",
                              "acquirer_sec_id", "acquirer_ous_id",
                              "announcement_date"],
                obj={
                    "target_sec_id": self.company_sec5.id,
                    "acquirer_sec_id": self.company_sec6.id,
                    "deal_value": 3000000,
                    "type": "Cash",
                    "annoucement_date": "02/14/2020",
                    "offer_price": 120,
                    "market_price": 105,
                    "dma_file_id": self.file3.id,
                },
            )
            self.assertEqual(
                MergerResourceSchema().dump(instance),
                {
                    "id": 2,
                    "target_sec_id": 3,
                    "target_ous_id": None,
                    "acquirer_sec_id": 4,
                    "acquirer_ous_id": None,
                    "deal_value": 3000000,
                    "type": "Cash",
                    "announcement_date": "2020-02-14T00:00:00+00:00",
                    "offer_price": 120.0,
                    "market_price": 105.0,
                    "dma_file_id": 2,
                    "updated_at": '2020-01-01T00:00:00+00:00',
                }
                ,
            )
            self.assertFalse(created)



from marshmallow import (
    Schema,
    fields,
    validate,
)


class TargetResourceSchema(Schema):
    not_blank = validate.Length(min=1, error='Field cannot be blank')

    id = fields.Integer(dump_only=True)
    cui = fields.String(required=True)
    name = fields.String(required=True)
    symbol = fields.String(required=True)
    bio_type = fields.String(allow_none=True)
    pathways = fields.String(allow_none=True)
    updated_at = fields.DateTime()



from marshmallow import (
    Schema,
    fields,
    validate,
    pre_load,
)

from ...utils.utils import convert_string_to_datetime


def _pre_load_datetime_fields(in_data):
    date_fields = [
        'decision_date',
        'petitioner_grant_date',
        'respondent_grant_date',
    ]

    for date_field in date_fields:
        if date_field in in_data:
            if(
                in_data[date_field] == '-' or
                in_data[date_field] is None
            ):
                in_data[date_field] = None
            else:
                in_data[date_field] = convert_string_to_datetime(
                    date=in_data[date_field],
                    string_format='%m-%d-%Y',
                )

    return in_data


class PTAB2DecisionResourceSchema(Schema):
    not_blank = validate.Length(min=1, error='Field cannot be blank')

    id = fields.Integer(dump_only=True)
    board_rulings = fields.String(allow_none=True)
    decision_date = fields.DateTime(allow_none=True)
    decision_type_category = fields.String(allow_none=True)
    document_identifier = fields.String(allow_none=True)
    document_name = fields.String(allow_none=True)
    identifier = fields.String(required=True)
    issue_type = fields.String(allow_none=True)
    petitioner_application_number_text = fields.String(allow_none=True)
    petitioner_counsel_name = fields.String(allow_none=True)
    petitioner_grant_date = fields.DateTime(allow_none=True)
    petitioner_group_art_unit_number = fields.String(allow_none=True)
    petitioner_inventor_name = fields.String(allow_none=True)
    petitioner_party_name = fields.String(allow_none=True)
    petitioner_patent_number = fields.String(allow_none=True)
    petitioner_patent_owner_name = fields.String(allow_none=True)
    petitioner_technology_center_number = fields.String(allow_none=True)
    proceeding_number = fields.String(required=True)
    proceeding_type_category = fields.String(allow_none=True)
    respondent_application_number_text = fields.String(allow_none=True)
    respondent_counsel_name = fields.String(allow_none=True)
    respondent_grant_date = fields.DateTime(allow_none=True)
    respondent_group_art_unit_number = fields.String(allow_none=True)
    respondent_inventor_name = fields.String(allow_none=True)
    respondent_party_name = fields.String(allow_none=True)
    respondent_patent_number = fields.String(allow_none=True)
    respondent_patent_owner_name = fields.String(allow_none=True)
    respondent_technology_center_number = fields.String(allow_none=True)
    subdecision_type_category = fields.String(allow_none=True)
    subproceeding_type_category = fields.String(allow_none=True)
    claim_invalidation_status = fields.String(allow_none=True)
    updated_at = fields.DateTime(dump_only=True)

    @pre_load
    def convert_string_to_datetime(self, in_data, **kwargs):
        return _pre_load_datetime_fields(in_data)


class PTAB2DecisionPatchSchema(Schema):
    not_blank = validate.Length(min=1, error='Field cannot be blank')

    id = fields.Integer(dump_only=True)
    board_rulings = fields.String(allow_none=True)
    decision_date = fields.DateTime(allow_none=True)
    decision_type_category = fields.String(allow_none=True)
    document_identifier = fields.String(allow_none=True)
    document_name = fields.String(allow_none=True)
    identifier = fields.String(required=True)
    issue_type = fields.String(allow_none=True)
    petitioner_application_number_text = fields.String(allow_none=True)
    petitioner_counsel_name = fields.String(allow_none=True)
    petitioner_grant_date = fields.DateTime(allow_none=True)
    petitioner_group_art_unit_number = fields.String(allow_none=True)
    petitioner_inventor_name = fields.String(allow_none=True)
    petitioner_party_name = fields.String(allow_none=True)
    petitioner_patent_number = fields.String(allow_none=True)
    petitioner_patent_owner_name = fields.String(allow_none=True)
    petitioner_technology_center_number = fields.String(allow_none=True)
    proceeding_number = fields.String(required=True)
    proceeding_type_category = fields.String(allow_none=True)
    respondent_application_number_text = fields.String(allow_none=True)
    respondent_counsel_name = fields.String(allow_none=True)
    respondent_grant_date = fields.DateTime(allow_none=True)
    respondent_group_art_unit_number = fields.String(allow_none=True)
    respondent_inventor_name = fields.String(allow_none=True)
    respondent_party_name = fields.String(allow_none=True)
    respondent_patent_number = fields.String(allow_none=True)
    respondent_patent_owner_name = fields.String(allow_none=True)
    respondent_technology_center_number = fields.String(allow_none=True)
    subdecision_type_category = fields.String(allow_none=True)
    subproceeding_type_category = fields.String(allow_none=True)
    claims_invalidation_status = fields.String(allow_none=True)


class PTAB2DecisionQueryParamsSchema(Schema):
    not_blank = validate.Length(min=1, error='Field cannot be blank')

    id = fields.Integer()
    identifier = fields.String()



from marshmallow import (
    Schema,
    fields,
    validate,
)


class NCTRefResourceSchema(Schema):
    not_blank = validate.Length(min=1, error="Field cannot be blank")

    id = fields.Integer(dump_only=True)
    nct_study_id = fields.Integer(required=True)
    pmid = fields.Integer()
    pubmed_id = fields.Integer(allow_none=True)
    reference_type = fields.String(allow_none=True)
    citation = fields.String()
    updated_at = fields.DateTime(dump_only=True)



import copy
from marshmallow import ValidationError
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import (
    NoResultFound,
)
from parameterized import parameterized

from test.backendtestcase import TestCase
from test.utils import second_equals_first
from src.cs_models.resources.Subsidiary import Subsidiary
from src.cs_models.resources.CompanySEC import CompanySEC
from src.cs_models.resources.CompanyOUS import CompanyOUS


class SubsidiaryResourceTestCase(TestCase):
    def setUp(self):
        super(SubsidiaryResourceTestCase, self).setUp()
        self.inst = Subsidiary()
        self.inst_sec_company = CompanySEC()
        self.inst_ous_company = CompanyOUS()

        self.company_sec1 = self.inst_sec_company.create({
            'cik_str': '200406',
            'ticker': 'JNJ',
            "title": "JOHNSON & JOHNSON",
        })

        self.company_sec2 = self.inst_sec_company.create({
            "cik_str": "816284",
            "ticker": "CELG",
            "title": "CELGENE CORP /DE/"
        })

        self.company_ous1 = self.inst_ous_company.create({
            'name': 'Aurobindo Pharma',
            'ticker': 'AUROPHARMA',
            'exchange': 'NSE',
        })

        self.subsidiary1 = self.inst.create({
            'name': 'Janssen Scientific Affairs, LLC',
            'company_sec_id': self.company_sec1['id'],
            'company_ous_id': None,
        })

        self.subsidiary2 = self.inst.create({
            'name': 'Celgene International II SARL',
            'company_sec_id': self.company_sec2['id'],
            'company_ous_id': None,
        })

        self.subsidiary3 = self.inst.create({
            'name': 'Aurobindo Pharma Pvt. Ltd.',
            'company_sec_id': None,
            'company_ous_id': self.company_ous1['id'],
        })

        self.valid_data = {
            'name': 'Janssen Research & Development, LLC',
            'company_sec_id': self.company_sec1['id'],
            'company_ous_id': None,
        }

    @parameterized.expand([
        ('name',),
    ])
    def test_create_validation_error_missing_field(self, field_to_pop):
        base_data = copy.copy(self.valid_data)
        base_data.pop(field_to_pop)
        self.assertRaises(
            ValidationError,
            self.inst.create,
            base_data,
        )

    def test_create_violates_unique_constraint(self):
        self.assertRaises(
            IntegrityError,
            self.inst.create,
            {'name': self.subsidiary1['name']},
        )

    def test_create(self):
        resp = self.inst.create(self.valid_data)
        second_equals_first(self.valid_data, resp)

    def test_read(self):
        resp = self.inst.read({})
        self.assertEqual(3, len(resp))

    @parameterized.expand([
        ('id', 'subsidiary1', 'id', 1),
        ('name', 'subsidiary1', 'name', 1),
    ])
    def test_read_w_params(
        self,
        field_name,
        attr,
        attr_field,
        expected_length,
    ):
        resp = self.inst.read({})
        self.assertEqual(3, len(resp))

        resp = self.inst.read({
            field_name: getattr(self, attr)[attr_field],
        })
        self.assertEqual(expected_length, len(resp))

    @parameterized.expand([
        ('id', 999, NoResultFound),
    ])
    def test_one_raises_exception(self, field_name, field_value, exception):
        self.assertRaises(
            exception,
            self.inst.one,
            {
                field_name: field_value,
            },
        )

    @parameterized.expand([
        ('id',),
        ('name',),
    ])
    def test_one(self, field_name):
        resp = self.inst.one({
            field_name: self.subsidiary1[field_name],
        })
        second_equals_first(
            self.subsidiary1,
            resp,
        )

    def test_upsert_validation_error(self):
        self.assertRaises(
            ValidationError,
            self.inst.upsert,
            {
                'company_sec_id': self.valid_data['company_sec_id'],
            }
        )

    def test_upsert_both_company_sec_company_ous_validation_error(self):
        data = {
            **self.valid_data,
            **{
               'company_ous_id': self.company_ous1['id'],
               },
        }
        self.assertRaises(
            ValidationError,
            self.inst.upsert,
            data
        )

    def test_upsert_creates_new_entry(self):
        data = copy.copy(self.valid_data)
        self.assertEqual(3, len(self.inst.read({})))
        self.inst.upsert(data)
        self.assertEqual(4, len(self.inst.read({})))

    def test_upsert_updates_existing_row(self):
        data = {
            **self.valid_data,
            **{'name': self.subsidiary1['name'],
               'company_sec_id': self.company_sec1['id'],
               },
        }
        resp = self.inst.upsert(data)
        second_equals_first(
            data,
            resp,
        )
        self.assertEqual(3, len(self.inst.read({})))

    def test_upsert_with_invalid_company_sec_id(self):
        invalid_id = 9999999
        data = {
            **self.valid_data,
            **{'name': self.subsidiary1['name'],
               'company_sec_id': invalid_id,
               },
        }
        self.assertRaises(
            IntegrityError,
            self.inst.upsert,
            data,
        )



from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm.exc import NoResultFound
from marshmallow import ValidationError, EXCLUDE

from ...database import db_session
from ...resources.SubsidiarySponsorMapping.models import (
    SubsidiarySponsorMappingModel,
)
from .schemas import (
    SubsidiarySponsorMappingResourceSchema,
    SubsidiarySponsorMappingQueryParamsSchema,
)
from ...utils.utils import update_model


schema_resource = SubsidiarySponsorMappingResourceSchema()
schema_params = SubsidiarySponsorMappingQueryParamsSchema()


class SubsidiarySponsorMapping(object):
    @staticmethod
    def create(params):
        """
        Args:
            params: dict(SubsidiarySponsorMappingResourceSchema)

        Returns: SubsidiarySponsorMappingResourceSchema

        Raises:
            ValidationError
            SQLAlchemyError
        """
        data, errors = schema_resource.load(params)
        if errors:
            raise ValidationError(errors)
        response = _helper_create(data)
        return response

    @staticmethod
    def read(params):
        """
        Args:
            params: dict(SubsidiarySponsorMappingQueryParamsSchema)

        Returns: List<SubsidiarySponsorMappingResourceSchema>

        Raises:
            ValidationError
        """
        data = schema_params.load(params, unknown=EXCLUDE)

        subsidiary_sponsor_mapping_query = _build_query(params=data)
        response = schema_resource.dump(
            subsidiary_sponsor_mapping_query, many=True)
        return response

    @staticmethod
    def one(params):
        """
        Args:
            params: dict(SubsidiarySponsorMappingQueryParamsSchema)

        Returns: SubsidiarySponsorMappingResourceSchema

        Raises:
            ValidationError
        """
        data = schema_params.load(params, unknown=EXCLUDE)

        subsidiary_sponsor_mapping_query = _build_query(params=data).one()
        response = schema_resource.dump(subsidiary_sponsor_mapping_query)
        return response

    @staticmethod
    def upsert(params):
        """
        Args:
            params: SubsidiarySponsorMappingResourceSchema

        Returns:
            SubsidiarySponsorMappingResourceSchema

        Raises:
            ValidationError
        """
        data, errors = schema_resource.load(params)
        if errors:
            raise ValidationError(errors)

        try:
            query_params = {
                'sponsor_id': params['sponsor_id'],
                'sponsor_name': params['sponsor_name'],
                'subsidiary_id': params['subsidiary_id'],
            }
            subsidiary_sponsor_mapping_query = _build_query(query_params).one()
            response = _helper_update(data, subsidiary_sponsor_mapping_query)
        except NoResultFound:
            response = _helper_create(data)
        return response


def _helper_create(data):
    new_subsidiary_sponsor_map = SubsidiarySponsorMappingModel(**data)
    try:
        db_session.add(new_subsidiary_sponsor_map)
        db_session.commit()
        subsidiary_sponsor_map_query = db_session.query(
            SubsidiarySponsorMappingModel
        ).get(
            new_subsidiary_sponsor_map.id
        )
        response = schema_resource.dump(
            subsidiary_sponsor_map_query
        )
        return response
    except SQLAlchemyError:
        db_session.rollback()
        raise


def _helper_update(data, subsidiary_sponsor_map_query):
    data['id'] = subsidiary_sponsor_map_query.id
    try:
        update_model(data, subsidiary_sponsor_map_query)
        db_session.commit()
        response = schema_resource.dump(subsidiary_sponsor_map_query)
        return response
    except SQLAlchemyError:
        db_session.rollback()
        raise


def _build_query(params):
    q = db_session.query(
        SubsidiarySponsorMappingModel)
    if params.get('id'):
        q = q.filter_by(id=params.get('id'))
    if params.get('sponsor_id'):
        q = q.filter_by(
            sponsor_id=params.get('sponsor_id'))
    if params.get('subsidiary_id'):
        q = q.filter_by(
            subsidiary_id=params.get('subsidiary_id'))
    return q



from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm.exc import NoResultFound
from marshmallow import ValidationError, EXCLUDE

from cs_models.utils.utils import update_model
from ...database import db_session
from ...resources.Config.models import (
    ConfigModel,
)
from .schemas import (
    ConfigResourceSchema,
    ConfigQueryParamsSchema,
    ConfigPatchSchema,
)


schema_resource = ConfigResourceSchema()
schema_params = ConfigQueryParamsSchema()
schema_patch = ConfigPatchSchema()


class Config(object):
    @staticmethod
    def create(params):
        """
        Args:
            params: dict(ConfigResourceSchema)

        Returns: ConfigResourceSchema

        Raises:
            ValidationError
            SQLAlchemyError
        """
        data = schema_resource.load(params, unknown=EXCLUDE)
        response = _helper_create(data)
        return response

    @staticmethod
    def read(params):
        """
        Args:
            params: dict(ConfigQueryParamsSchema)

        Returns: List<ConfigResourceSchema>

        Raises:
            ValidationError
        """
        data = schema_params.load(params, unknown=EXCLUDE)
        config_query = _build_query(params=data)
        response = schema_resource.dump(
            config_query, many=True)
        return response

    @staticmethod
    def one(params):
        """
        Args:
            params: dict(ConfigQueryParamsSchema)

        Returns: ConfigResourceSchema

        Raises:
            ValidationError
        """
        data = schema_params.load(params, unknown=EXCLUDE)

        config_query = _build_query(params=data).one()
        response = schema_resource.dump(config_query)
        return response

    @staticmethod
    def update(id, params):
        """

        Args:
            id: int
            params: ConfigPatchSchema

        Returns:
            ConfigResourceSchema

        Raises:
            sqlalchemy.orm.exc.NoResultFound
            sqlalchemy.orm.exc.MultipleResultsFound
            ValidationError
        """
        config_query = db_session.query(
            ConfigModel).filter_by(id=id).one()
        data = schema_patch.load(params, unknown=EXCLUDE)

        response = _helper_update(data, config_query)
        return response

    @staticmethod
    def upsert(params):
        """
        Args:
            params: ConfigResourceSchema

        Returns:
            NewswireResourceSchema

        Raises:
            ValidationError
        """
        data = schema_resource.load(params, unknown=EXCLUDE)

        try:
            query_params = {
                'table_name': params['table_name'],
                'field_name': params['field_name'],
            }
            config_query = _build_query(query_params).one()
            response = _helper_update(data, config_query)
        except NoResultFound:
            response = _helper_create(data)
        return response

    @staticmethod
    def delete(id):
        """
        Args:
            id: int

        Returns: string

        Raises:
            sqlalchemy.orm.exc.NoResultFound
            sqlalchemy.orm.exc.MultipleResultsFound
            SQLAlchemyError
        """
        config_query = db_session.query(
            ConfigModel).filter_by(id=id).one()
        try:
            db_session.delete(config_query)
            db_session.commit()
            db_session.close()
            return 'Successfully deleted'
        except SQLAlchemyError:
            db_session.rollback()
            db_session.close()
            raise


def _helper_create(data):
    new_config = ConfigModel(**data)
    try:
        db_session.add(new_config)
        db_session.commit()
        config_query = db_session.query(
            ConfigModel).get(
            new_config.id)
        response = schema_resource.dump(config_query)
        db_session.close()
        return response
    except SQLAlchemyError:
        db_session.rollback()
        db_session.close()
        raise


def _helper_update(data, config_query):
    data['id'] = config_query.id
    try:
        update_model(data, config_query)
        db_session.commit()
        response = schema_resource.dump(config_query)
        return response
    except SQLAlchemyError:
        db_session.rollback()
        raise


def _build_query(params):
    q = db_session.query(
        ConfigModel)
    if params.get('id'):
        q = q.filter_by(id=params.get('id'))
    if params.get('table_name'):
        q = q.filter_by(table_name=params.get('table_name'))
    if params.get('field_name'):
        q = q.filter_by(field_name=params.get('field_name'))
    return q



from datetime import datetime

from sqlalchemy import (
    Column,
    Integer,
    DateTime,
    Boolean,
    ForeignKey,
)

from ...database import Base


class CompanyLicensingCollabModel(Base):
    __tablename__ = "company_licensing_collab"

    id = Column(Integer, primary_key=True)
    licensing_collab_id = Column(
        Integer,
        ForeignKey('licensing_collab.id'),
        nullable=False,
    )
    company_sec_id = Column(
        Integer,
        ForeignKey('companies_sec.id'),
        nullable=True,
    )
    company_ous_id = Column(
        Integer,
        ForeignKey('companies_ous.id'),
        nullable=True,
    )
    target_flag = Column(
        Boolean,
        nullable=True,
    )
    updated_at = Column(
        DateTime,
        nullable=False,
        # https://stackoverflow.com/questions/58776476/why-doesnt-freezegun-work-with-sqlalchemy-default-values
        default=lambda: datetime.utcnow(),
        onupdate=lambda: datetime.utcnow(),
    )



from datetime import datetime

from sqlalchemy import (
    Column,
    Integer,
    DateTime,
    UniqueConstraint,
    ForeignKey,
    String,
)

from ...database import Base


class HLTPTMapModel(Base):
    __tablename__ = "hlt_pt_map"

    id = Column(Integer, primary_key=True)
    hlt_id = Column(
        Integer,
        ForeignKey('hlt_terms.id'),
        nullable=False,
    )
    pt_code = Column(String(128), nullable=False, index=True)
    updated_at = Column(
        DateTime,
        nullable=False,
        # https://stackoverflow.com/questions/58776476/why-doesnt-freezegun-work-with-sqlalchemy-default-values
        default=lambda: datetime.utcnow(),
        onupdate=lambda: datetime.utcnow(),
    )

    __table_args__ = (UniqueConstraint("hlt_id", "pt_code"),)



from datetime import datetime
from sqlalchemy import (
    Column,
    Integer,
    DateTime,
    ForeignKey,
)

from ...database import Base


class ClaimDependencyModel(Base):
    __tablename__ = 'claim_dependencies'

    id = Column(Integer, primary_key=True)
    claim_id = Column(
        Integer,
        ForeignKey('claims.id'),
        nullable=False,
    )
    parent_claim_id = Column(
        Integer,
        ForeignKey('claims.id'),
        nullable=False,
    )
    updated_at = Column(
        DateTime,
        nullable=False,
        # https://stackoverflow.com/questions/58776476/why-doesnt-freezegun-work-with-sqlalchemy-default-values
        default=lambda: datetime.utcnow(),
        onupdate=lambda: datetime.utcnow(),
    )



from datetime import datetime

from sqlalchemy import (
    Column,
    Integer,
    String,
    Text,
    Float,
    DateTime,
    ForeignKey,
)

from ...database import Base


class ConditionEpidemiologyModel(Base):
    __tablename__ = "condition_epidemiology"

    id = Column(Integer, primary_key=True)
    source_date = Column(DateTime, nullable=False)
    source_title = Column(Text, nullable=False)
    source_name = Column(String(128), nullable=False)
    source_table = Column(String(50), nullable=False)
    source_table_id = Column(Integer, nullable=False)
    pubmed_id = Column(
        Integer,
        ForeignKey('pubmed.id'),
        nullable=True,
    )
    news_id = Column(
        Integer,
        ForeignKey('newswires.id'),
        nullable=True,
    )
    measure = Column(String(50), nullable=False)
    geographic_area = Column(String(50))
    snippet = Column(Text, nullable=False)
    statistic = Column(Float)
    statistic_text = Column(String(128))
    year = Column(String(50))
    population_type = Column(String(128))
    updated_at = Column(
        DateTime,
        nullable=False,
        # https://stackoverflow.com/questions/58776476/why-doesnt-freezegun-work-with-sqlalchemy-default-values
        default=lambda: datetime.utcnow(),
        onupdate=lambda: datetime.utcnow(),
    )



# flake8: noqa

from ...cs_models.resources.Claim.models import ClaimModel
from ...cs_models.resources.ClaimChallenged.models import (
    ClaimChallengedModel,
)
from ...cs_models.resources.ClaimDependency.models import (
    ClaimDependencyModel,
)
from ...cs_models.resources.Company.models import CompanyModel
from ...cs_models.resources.CrossRef.models import CrossRefModel
from ...cs_models.resources.DocumentReference.models import (
    DocumentReferenceModel,
)
from ...cs_models.resources.File.models import FileModel
from ...cs_models.resources.InstitutionProbability.models import (
    InstitutionProbabilityModel,
)
from ...cs_models.resources.OrangeBookPatent.models import (
    OrangeBookPatentModel,
)
from ...cs_models.resources.OrangeBookProduct.models import (
    OrangeBookProductModel,
)
from ...cs_models.resources.PacerCase.models import (
    PacerCaseModel,
)
from ...cs_models.resources.Patent.models import (
    PatentModel,
)
from ...cs_models.resources.PatentApplication.models import (
    PatentApplicationModel,
)
from ...cs_models.resources.PatentCompound.models import (
    PatentCompoundModel,
)
from ...cs_models.resources.PatentIdentifierSynonym.models import (
    PatentIdentifierSynonymModel,
)
from ...cs_models.resources.PriorArt.models import (
    PriorArtModel,
)
from ...cs_models.resources.PriorArtRefRelation.models import (
    PriorArtRefRelationModel,
)
from ...cs_models.resources.PTAB2Decision.models import (
    PTAB2DecisionModel,
)
from ...cs_models.resources.PTAB2Document.models import (
    PTAB2DocumentModel,
)
from ...cs_models.resources.PTAB2Proceeding.models import (
    PTAB2ProceedingModel,
)
from ...cs_models.resources.RelatedMatter.models import (
    RelatedMatterModel,
)
from ...cs_models.resources.TrialStat.models import (
    TrialStatModel,
)
from ...cs_models.resources.UserSearchHistoryItem.models import (
    UserSearchHistoryItemModel,
)



from marshmallow import ValidationError
from sqlalchemy.orm.exc import (
    NoResultFound,
    MultipleResultsFound,
)

from test.backendtestcase import TestCase
from test.utils import second_equals_first
from src.cs_models.resources.PatentCompound import (
    PatentCompound,
)
from src.cs_models.resources.Patent import Patent
from src.cs_models.resources.File import File


class PatentCompoundResourceTestCase(TestCase):
    def setUp(self):
        super(PatentCompoundResourceTestCase, self).setUp()
        self.inst_patent = Patent()
        self.inst_file = File()
        self.inst = PatentCompound()
        self.file1 = self.inst_file.create({
            'file_format': 'txt',
            's3_bucket_name': 'test-bucket',
            's3_key_name': 'key1',
        })
        self.patent1 = self.inst_patent.create({
            'patent_number': '1000',
            'app_grp_art_number': '5000',
            'jurisdiction': 'USPAT',
        })
        self.patent2 = self.inst_patent.create({
            'patent_number': '2000',
            'app_grp_art_number': '5000',
            'jurisdiction': 'USPAT',
        })
        self.patent_compound1 = self.inst.create({
            'patent_id': self.patent1['id'],
            'cid': 'some cid',
            'compound_name': 'some compound_name',
            'iupac_name': 'some iupac_name',
            'molecular_formula': 'some molecular_formula',
            'molecular_weight': 'some molecular_weight',
            'structure_file_id': self.file1['id'],
        })

    def test_create_validation_error_missing_fields(self):
        self.assertRaises(
            ValidationError,
            self.inst.create,
            {},
        )

    def test_create_validation_error_blank_cid(self):
        self.assertRaises(
            ValidationError,
            self.inst.create,
            {
                'patent_id': 123,
                'cid': '',
                'compound_name': 'some compound_name',
                'iupac_name': 'some iupac_name',
                'molecular_formula': 'some molecular_formula',
                'molecular_weight': 'some molecular_weight',
            },
        )

    def test_create_duplicate_patent_id(self):
        self.assertRaises(
            Exception,
            self.inst.create,
            {
                'patent_id': self.patent_compound1['patent_id'],
                'cid': self.patent_compound1['cid'],
                'compound_name': 'some compound_name',
                'iupac_name': 'some iupac_name',
                'molecular_formula': 'some molecular_formula',
                'molecular_weight': 'some molecular_weight',
            }
        )

    def test_create(self):
        new_data = {
            'patent_id': self.patent2['id'],
            'cid': 'some cid',
            'compound_name': 'some compound_name',
            'iupac_name': 'some iupac_name',
            'molecular_formula': 'some molecular_formula',
            'molecular_weight': 'some molecular_weight',
            'structure_file_id': self.file1['id'],
        }
        response = self.inst.create(new_data)
        second_equals_first(new_data, response)

    def test_read_validation_error_blank_cid(self):
        self.assertRaises(
            ValidationError,
            self.inst.read,
            {'cid': ''},
        )

    def test_read_all(self):
        response = self.inst.read({})
        self.assertEqual(len(response), 1)

    def test_read_w_params(self):
        # setup
        self.inst.create({
            'patent_id': self.patent2['id'],
            'cid': 'some cid2',
            'compound_name': 'some compound_name2',
            'iupac_name': 'some iupac_name2',
            'molecular_formula': 'some molecular_formula',
            'molecular_weight': 'some molecular_weight',
            'structure_file_id': self.file1['id'],
        })

        response = self.inst.read({})
        self.assertEqual(len(response), 2)

        response = self.inst.read({'patent_id': self.patent1['id']})
        self.assertEqual(len(response), 1)

        response = self.inst.read({'cid': self.patent_compound1['cid']})
        self.assertEqual(len(response), 1)

        response = self.inst.read({
            'compound_name': self.patent_compound1['compound_name'],
        })
        self.assertEqual(len(response), 1)

        response = self.inst.read({
            'iupac_name': self.patent_compound1['iupac_name'],
        })
        self.assertEqual(len(response), 1)

    def test_one_raises_MultipleResultsFound(self):
        # setup
        self.inst.create({
            'patent_id': self.patent2['id'],
            'cid': 'some cid',
            'compound_name': 'some compound_name2',
            'iupac_name': 'some iupac_name2',
            'molecular_formula': 'some molecular_formula',
            'molecular_weight': 'some molecular_weight',
            'structure_file_id': self.file1['id'],
        })

        self.assertRaises(
            MultipleResultsFound,
            self.inst.one,
            {'cid': self.patent_compound1['cid']},
        )

    def test_one_raises_NoResultFound(self):
        self.assertRaises(
            NoResultFound,
            self.inst.one,
            {'cid': '123123'},
        )



from marshmallow import (
    Schema,
    fields,
    validate,
)


class MergerActionRecordeResourceSchema(Schema):
    not_blank = validate.Length(min=1, error='Field cannot be blank')

    id = fields.Integer(dump_only=True)
    merger_action_id = fields.Integer(required=True)
    table = fields.String(required=True)
    record_id = fields.Integer(required=True)
    updated_at = fields.DateTime()



from sqlalchemy import (
    Column,
    Integer,
    String,
    Text,
    DateTime,
    ForeignKey,
    UniqueConstraint,
)
from datetime import datetime
from ...database import Base


class PriorArtModel(Base):
    __tablename__ = 'prior_arts'

    id = Column(Integer, primary_key=True)
    ptab2_document_id = Column(
        Integer,
        ForeignKey('ptab2_documents.id'),
    )
    tag = Column(String(128), nullable=False)
    title = Column(Text)
    exhibit = Column(String(128))
    updated_at = Column(
        DateTime,
        nullable=False,
        # https://stackoverflow.com/questions/58776476/why-doesnt-freezegun-work-with-sqlalchemy-default-values
        default=lambda: datetime.utcnow(),
        onupdate=lambda: datetime.utcnow(),
    )

    __table_args__ = (UniqueConstraint('ptab2_document_id', 'tag'),)



from marshmallow import (
    Schema,
    fields,
    validate,
)


class PriorArtResourceSchema(Schema):
    not_blank = validate.Length(min=1, error='Field cannot be blank')

    id = fields.Integer(dump_only=True)
    ptab2_document_id = fields.Integer(required=True)
    tag = fields.String(required=True, validate=not_blank)
    title = fields.String(required=True, allow_none=True)
    exhibit = fields.String(required=True, allow_none=True)
    updated_at = fields.DateTime()


class PriorArtQueryParamsSchema(Schema):
    not_blank = validate.Length(min=1, error='Field cannot be blank')

    id = fields.Integer()
    ptab2_document_id = fields.Integer()
    tag = fields.String(validate=not_blank)


class PriorArtPatchSchema(Schema):
    not_blank = validate.Length(min=1, error='Field cannot be blank')

    ptab2_document_id = fields.Integer()
    tag = fields.String()
    title = fields.String()
    exhibit = fields.String()
    updated_at = fields.DateTime()



from sqlalchemy import (
    Column,
    Integer,
    DateTime,
    ForeignKey,
    UniqueConstraint,
)
from datetime import datetime

from ...database import Base


class PriorArtRefRelationModel(Base):
    __tablename__ = 'prior_art_ref_relations'

    id = Column(Integer, primary_key=True)
    prior_art_id = Column(
        Integer,
        ForeignKey('prior_arts.id'),
        nullable=False,
    )
    cross_ref_id = Column(
        Integer,
        ForeignKey('cross_refs.id'),
    )
    patent_id = Column(
        Integer,
        ForeignKey('patents.id'),
    )
    patent_application_id = Column(
        Integer,
        ForeignKey('patent_applications.id'),
    )
    updated_at = Column(
        DateTime,
        nullable=False,
        default=datetime.utcnow,
        onupdate=datetime.utcnow,
    )

    __table_args__ = (
        UniqueConstraint('prior_art_id', 'cross_ref_id'),
        UniqueConstraint('prior_art_id', 'patent_id'),
        UniqueConstraint('prior_art_id', 'patent_application_id'),
    )



from sqlalchemy.exc import SQLAlchemyError
from marshmallow import ValidationError, EXCLUDE

from ...database import db_session
from ...resources.NDCSetIDCross.models import NDCSetIDCrossModel
from .schemas import (
    NDCSetIDCrossPatchSchema,
    NDCSetIDCrossQueryParamsSchema,
    NDCSetIDCrossResourceSchema,
)


schema_resource = NDCSetIDCrossResourceSchema()
schema_params = NDCSetIDCrossQueryParamsSchema()
schema_patch = NDCSetIDCrossPatchSchema()


class NDCSetIDCross(object):
    @staticmethod
    def create(params):
        """
        Args:
            params: dict(NDCSetIDCrossResourceSchema)

        Returns: NDCSetIDCrossResourceSchema

        Raises:
            ValidationError
            SQLAlchemyError
        """
        data = schema_resource.load(params, unknown=EXCLUDE)
        response = _helper_create(data)
        return response

    @staticmethod
    def read(params):
        """
        Args:
            params: dict(NDCSetIDCrossQueryParamsSchema)

        Returns: List<NDCSetIDCrossResourceSchema>

        Raises:
            ValidationError
        """
        data = schema_params.load(params, unknown=EXCLUDE)

        ndc_setid_cross_query = _build_query(params=data)
        response = schema_resource.dump(ndc_setid_cross_query, many=True)
        return response

    @staticmethod
    def one(params):
        """
        Args:
            params: dict(NDCSetIDCrossQueryParamsSchema)

        Returns: NDCSetIDCrossResourceSchema

        Raises:
            ValidationError
        """
        data = schema_params.load(params, unknown=EXCLUDE)

        ndc_setid_cross_query = _build_query(params=data).one()
        response = schema_resource.dump(ndc_setid_cross_query)
        return response


def _helper_create(data):
    new_ndc_setid_cross = NDCSetIDCrossModel(**data)
    try:
        db_session.add(new_ndc_setid_cross)
        db_session.commit()
        druglabel_query = db_session.query(NDCSetIDCrossModel).get(
            new_ndc_setid_cross.product_ndc)
        response = schema_resource.dump(druglabel_query)
        db_session.close()
        return response
    except SQLAlchemyError:
        db_session.rollback()
        db_session.close()
        raise


def _build_query(params):
    q = db_session.query(NDCSetIDCrossModel)
    if params.get("product_ndc"):
        q = q.filter_by(product_ndc=params.get("product_ndc"))
    if params.get("spl_set_id"):
        q = q.filter_by(spl_set_id=params.get("spl_set_id"))
    if params.get("package_ndc"):
        q = q.filter_by(package_ndc=params.get("package_ndc"))
    return q



from datetime import datetime

from sqlalchemy import (
    Column,
    Integer,
    DateTime,
    Text,
    ForeignKey,
)

from ...database import Base


class SalesModel(Base):
    __tablename__ = "sales"

    id = Column(Integer, primary_key=True)
    sec_filing_id = Column(
        Integer,
        ForeignKey('companies_sec_filings.id'),
        nullable=False,
    )
    table_html_file_id = Column(
        Integer,
        ForeignKey('files.id'),
        nullable=False,
    )
    table_number = Column(Integer, nullable=False)
    table_preceding_info = Column(Text, nullable=True)
    updated_at = Column(
        DateTime,
        nullable=False,
        # https://stackoverflow.com/questions/58776476/why-doesnt-freezegun-work-with-sqlalchemy-default-values
        default=lambda: datetime.utcnow(),
        onupdate=lambda: datetime.utcnow(),
    )




from marshmallow import (
    Schema,
    fields,
    pre_load,
    ValidationError,
)


class PriorArtRefRelationResourceSchema(Schema):
    id = fields.Integer(dump_only=True)
    prior_art_id = fields.Integer(required=True)
    cross_ref_id = fields.Integer()
    patent_id = fields.Integer()
    patent_application_id = fields.Integer()
    updated_at = fields.DateTime(dump_only=True)

    @pre_load
    def check_ref_fields(self, in_data, **kwargs):
        if self._get_number_of_ref_fields(in_data) != 1:
            raise ValidationError('Provide only one of the ref fields')
        return in_data

    def _get_number_of_ref_fields(self, in_data, **kwargs):
        result = 0
        if 'cross_ref_id' in in_data:
            result += 1
        if 'patent_id' in in_data:
            result += 1
        if 'patent_application_id' in in_data:
            result += 1
        return result


class PriorArtRefRelationQueryParamsSchema(Schema):
    id = fields.Integer()
    prior_art_id = fields.Integer()
    cross_ref_id = fields.Integer()
    patent_id = fields.Integer()
    patent_application_id = fields.Integer()



from marshmallow import (
    Schema,
    fields,
    validate,
    pre_load,
    ValidationError,
)
from ...utils.utils import pre_load_date_fields


class StreamResourceSchema(Schema):
    not_blank = validate.Length(min=1, error="Field cannot be blank")

    id = fields.Integer(dump_only=True)
    date = fields.DateTime(required=True)
    resource_type = fields.String(required=True)
    intervention_id = fields.Integer(allow_none=True)
    condition_id = fields.Integer(allow_none=True)
    company_sec_id = fields.Integer(allow_none=True)
    company_ous_id = fields.Integer(allow_none=True)
    ptab2_proceeding_id = fields.Integer(allow_none=True)
    source_table = fields.String(required=True)
    source_table_id = fields.Integer(required=True)
    news_id = fields.Integer(allow_none=True)
    sec_filing_id = fields.Integer(allow_none=True)
    ptab2_document_id = fields.Integer(allow_none=True)
    pubmed_id = fields.Integer(allow_none=True)
    company_filing_id = fields.Integer(allow_none=True)
    updated_at = fields.DateTime(dump_only=True)

    @pre_load
    def check_resource_info(self, in_data, **kwargs):
        if self._get_number_of_resource_fields(in_data) == 0:
            raise ValidationError('Provide at least one resource info')
        if self._get_number_of_source_fields(in_data) != 1:
            raise ValidationError('Provide only one source info')
        return in_data

    def _get_number_of_resource_fields(self, in_data, **kwargs):
        result = 0
        if 'intervention_id' in in_data:
            if in_data['intervention_id'] is not None:
                result += 1
        if 'condition_id' in in_data:
            if in_data['condition_id'] is not None:
                result += 1
        if 'company_sec_id' in in_data:
            if in_data['company_sec_id'] is not None:
                result += 1
        if 'company_ous_id' in in_data:
            if in_data['company_ous_id'] is not None:
                result += 1
        if 'ptab2_proceeding_id' in in_data:
            if in_data['ptab2_proceeding_id'] is not None:
                result += 1
        return result

    def _get_number_of_source_fields(self, in_data, **kwargs):
        result = 0
        if 'news_id' in in_data:
            if in_data['news_id'] is not None:
                result += 1
        if 'sec_filing_id' in in_data:
            if in_data['sec_filing_id'] is not None:
                result += 1
        if 'ptab2_document_id' in in_data:
            if in_data['ptab2_document_id'] is not None:
                result += 1
        if 'pubmed_id' in in_data:
            if in_data['pubmed_id'] is not None:
                result += 1
        if 'company_filing_id' in in_data:
            if in_data['company_filing_id'] is not None:
                result += 1
        return result

    @pre_load
    def convert_string_to_datetime(self, in_data, **kwargs):
        date_fields = ['date']

        in_data = pre_load_date_fields(
            in_data,
            date_fields,
            date_format="%Y%m%dT%H%M%S",
        )
        return in_data



import os
from neo4j import GraphDatabase

uri = os.getenv("GRAPH_DB_URI", "bolt://localhost:7687")
user = os.getenv("GRAPH_DB_USER", "neo4j")
password = os.getenv("GRAPH_DB_PASSWORD", "mindgram12")


def create_driver(uri, user, password):

    def resolver(address):
        host, port = address
        if host == "x.example.com":
            yield "a.example.com", port
            yield "b.example.com", port
            yield "c.example.com", port
        else:
            yield host, port

    return GraphDatabase.driver(uri, auth=(user, password), resolver=resolver)


driver = create_driver(uri, user, password)



from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.sql.operators import in_op
from datetime import datetime
from marshmallow import ValidationError, EXCLUDE

from ...database import db_session
from ...utils.utils import update_model
from ...resources.OrangeBookPatent.models import (
    OrangeBookPatentModel,
)
from .schemas import (
    OrangeBookPatentResourceSchema,
    OrangeBookPatentQueryParamsSchema,
    OrangeBookPatentPatchSchema,
)


schema_resource = OrangeBookPatentResourceSchema()
schema_params = OrangeBookPatentQueryParamsSchema()
schema_patch = OrangeBookPatentPatchSchema()


class DBException(SQLAlchemyError):
    pass


class OrangeBookPatentNotFoundException(Exception):
    def __init__(self, message):
        self.message = message


class OrangeBookPatent:
    @staticmethod
    def create(params):
        """
        :param
            Dict: OrangeBookPatentResourceSchema
        :return:
            OrangeBookPatentResourceSchema
        :exception:
            ValidationError
            DBException
        """
        data, errors = schema_resource.load(params)
        if errors:
            raise ValidationError(errors)

        response = _helper_create(data)
        return response

    @staticmethod
    def read(params):
        """
        :param
            params: OrangeBookPatentQueryParamsSchema
        :return:
            List<OrangeBookPatentResourceSchema>
        :exception:
            ValidationError
        """

        data = schema_params.load(params, unknown=EXCLUDE)

        orange_book_patent_query = _build_query(params=data)
        response = schema_resource.dump(
            orange_book_patent_query, many=True)
        return response

    @staticmethod
    def one(params):
        """
        :param
            params: OrangeBookPatentQueryParamsSchema
        :return:
            OrangeBookPatentResourceSchema
        :exception:
            ValidationError
        """

        data = schema_params.load(params, unknown=EXCLUDE)

        orange_book_patent_query = _build_query(params=data).one()
        response = schema_resource.dump(orange_book_patent_query)
        return response

    @staticmethod
    def update(id, params):
        """
        :param
            id: integer: required
            params: OrangeBookPatentPatchSchema
        :return:
            OrangeBookPatentResourceSchema
        :exception:
            OrangeBookPatentNotFoundException
            ValidationError
            DBException
        """
        orange_book_patent_query = db_session.query(
            OrangeBookPatentModel).filter_by(
            id=id,
        ).first()
        if not orange_book_patent_query:
            raise OrangeBookPatentNotFoundException('OB Patent not found!')
        data, errors = schema_patch.load(params)
        if errors:
            raise ValidationError(errors)

        response = _helper_update(data, orange_book_patent_query)
        return response


def _helper_create(data):
    new_orange_book_patent = OrangeBookPatentModel(
        updated_at=datetime.utcnow(),
        **data,
    )
    try:
        db_session.add(new_orange_book_patent)
        db_session.commit()
        orange_book_patent_query = db_session.query(
            OrangeBookPatentModel).get(
            new_orange_book_patent.id,
        )
        response = schema_resource.dump(orange_book_patent_query)
        db_session.close()
        return response
    except SQLAlchemyError:
        db_session.rollback()
        db_session.close()
        raise


def _helper_update(data, orange_book_patent_query):
    data['id'] = orange_book_patent_query.id
    data['updated_at'] = datetime.utcnow()
    try:
        update_model(data, orange_book_patent_query)
        db_session.commit()
        response = schema_resource.dump(orange_book_patent_query)
        return response
    except SQLAlchemyError:
        db_session.rollback()
        raise


def _build_query(params):
    q = db_session.query(
        OrangeBookPatentModel)
    if params.get('id'):
        q = q.filter_by(id=params.get('id'))
    if params.get('appl_type'):
        q = q.filter_by(appl_type=params.get('appl_type'))
    if params.get('appl_no'):
        q = q.filter_by(appl_no=params.get('appl_no'))
    if 'appl_nos' in params:
        q = q.filter(in_op(
            OrangeBookPatentModel.appl_no,
            params.get('appl_nos'),
        ))
    if params.get('product_no'):
        q = q.filter_by(product_no=params.get('product_no'))
    if params.get('patent_no'):
        q = q.filter_by(patent_no=params.get('patent_no'))
    if params.get('patent_use_code'):
        q = q.filter_by(patent_use_code=params.get('patent_use_code'))
    return q



# pylint: disable=duplicate-code
import os
import sshtunnel
import typing
from contextlib import contextmanager

from marshmallow import EXCLUDE
from sqlalchemy import event
from sqlalchemy.orm import Session

from ..aact_database import engine as engine_aact
from ..database import engine, get_engine_for_port

# Uncomment the logging statements to see the SQL
# queries that are made
# import logging
# logging.basicConfig()
# logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)


@contextmanager
def assert_sql(expected):
    stmts = []

    # pylint: disable=unused-argument
    def catch_queries(conn, cursor, statement, *args):
        stmts.append((statement, args[0]))

    event.listen(engine, "before_cursor_execute", catch_queries)

    yield

    if expected != stmts:
        raise AssertionError(f"Expected: {expected!r}, Got: {stmts!r}")


@contextmanager
def session_scope(db: typing.Optional[str] = None):
    """Provide a transactional scope around a
    series of operations."""
    # https://docs.sqlalchemy.org/en/13/orm/session_transaction.html
    if os.getenv("SSH_TUNNEL"):
        ssh_server = os.getenv("SSH_HOST")
        ssh_username = os.getenv("SSH_USERNAME")
        ssh_pkey_path = os.getenv("SSH_PKEY_PATH")
        remote_bind_address = os.getenv("MYSQL_HOST")
        with sshtunnel.SSHTunnelForwarder(
                (ssh_server, 22),
                ssh_username=ssh_username,
                ssh_pkey=ssh_pkey_path,
                remote_bind_address=(remote_bind_address, 3306),
        ) as tunnel:
            tunnel.start()
            bind_engine = get_engine_for_port(tunnel.local_bind_port)
            if db == "aact":
                bind_engine = engine_aact
            session = Session(bind=bind_engine)

            try:
                yield session
                session.commit()
            except Exception:
                session.rollback()
                raise
            finally:
                session.close()
    else:
        bind_engine = engine
        if db == "aact":
            bind_engine = engine_aact
        session = Session(bind=bind_engine)

        try:
            yield session
            session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()


def create(session, model, schema, obj):
    """Validates the `obj` against the `schema` and
    creates a new row."""
    data = schema().load(obj, unknown=EXCLUDE)

    instance = model(**data)
    session.add(instance)

    # https://stackoverflow.com/questions/4201455/sqlalchemy-whats-the-difference-between-flush-and-commit
    # Committing every small operation is not ideal but
    # is needed since we have multiple workers and we don't
    # want stale "ids". Had we used `session.flush`,
    # then different workers may clash when it comes
    # to generating ids
    session.commit()

    # Not doing `schema.dump` because it leads to a `SELECT ..`
    # query to db.
    return instance


def get_or_create(session, model, schema, obj: typing.Dict) -> typing.Tuple:
    """Based on the `obj`, gets the corresponding row
    if it exists. If it doesn't exist, then creates
    a new row."""
    instance = session.query(model).filter_by(**obj).first()
    if instance:
        return instance, False

    data = schema().load(obj, unknown=EXCLUDE)

    instance = model(**data)
    session.add(instance)
    session.commit()
    return instance, True


def get_or_create_by_filter(
    session, model, schema, obj: typing.Dict, query_fields: typing.List[str]
) -> typing.Tuple:
    """Builds a "query filter" based on `obj` and `query_fields`. If a row
    exists for the "query filter", gets the corresponding row
    if it exists. If it doesn't exist, then creates
    a new row."""
    # Attempt to get the model instance based on the
    # `query_fields` in `obj`
    instance = (
        session.query(model)
        .filter_by(
            **{
                key: obj[key]
                for key in obj
                if key in query_fields or key == "id"
            }
        )
        .first()
    )
    if instance:
        return instance, False

    data = schema().load(obj, unknown=EXCLUDE)

    instance = model(**data)
    session.add(instance)
    session.commit()
    return instance, True


def update_or_create(
    session, model, schema, obj: typing.Dict, query_fields: typing.List[str]
) -> typing.Tuple:
    """Builds a "query filter" based on `obj` and `query_fields`. If a row
    exists for the "query filter" then updates the row with the `obj` values.
    If a row doesn't exist, then creates a new row with the
    values from `obj`."""
    # Attempt to get the model instance based on the
    # `query_fields` in `obj`
    instance = (
        session.query(model)
        .filter_by(
            **{
                key: obj[key]
                for key in obj
                if key in query_fields or key == "id"
            }
        )
        .first()
    )

    # Validate
    data = schema().load(obj, unknown=EXCLUDE)

    if instance:
        # Update
        for k, v in data.items():
            instance.__setattr__(k, v)
        session.commit()
        return instance, False

    # Create
    instance = model(**data)
    session.add(instance)
    session.commit()
    return instance, True


def bulk_save(session, instances: typing.List, batch_size: int = 5000):
    """
    Updates or Creates the model instances that are passed in.
    Note: The validation of the data should be done by the caller before
    preparing the model instances.

    https://stackoverflow.com/questions/19904176/transactions-and-sqlalchemy
    https://docs.sqlalchemy.org/en/13/_modules/examples/performance/bulk_inserts.html
    https://docs.sqlalchemy.org/en/13/faq/performance.html
    """
    # Need to do this to avoid `sqlalchemy.orm.exc.StaleDataError`
    # when some `instances` involve updates and not creates.
    # Ends up performing extra `SELECT` queries.
    session.flush()

    for i in range(0, len(instances), batch_size):
        batch = instances[i : i + batch_size]
        session.bulk_save_objects(batch)
    session.commit()



from marshmallow import (
    Schema,
    fields,
    validate,
    pre_load,
)
from ...utils.utils import pre_load_date_fields


class InterventionMilestoneResourceSchema(Schema):
    not_blank = validate.Length(min=1, error='Field cannot be blank')

    id = fields.Integer(dump_only=True)
    intervention_condition_id = fields.Integer(required=True)
    start_date = fields.DateTime(required=True)
    end_date = fields.DateTime(required=True)
    milestone = fields.String(validate=not_blank, required=True)
    description = fields.String(allow_none=True)
    source_url = fields.String(allow_none=True)
    news_id = fields.Integer(allow_none=True)
    sec_filing_id = fields.Integer(allow_none=True)
    is_deleted = fields.Boolean(allow_none=True)
    updated_at = fields.DateTime()

    @pre_load
    def convert_string_to_datetime(self, in_data, **kwargs):
        date_fields = ['start_date', 'end_date']

        in_data = pre_load_date_fields(
            in_data,
            date_fields,
            date_format="%Y%m%dT%H%M%S",
        )
        return in_data


class InterventionMilestonePatchSchema(Schema):
    not_blank = validate.Length(min=1, error='Field cannot be blank')
    id = fields.Integer(required=True)
    is_deleted = fields.Boolean(allow_none=True)
    updated_at = fields.DateTime()



from marshmallow import (
    Schema,
    fields,
    validate,
)


class DocTextResourceSchema(Schema):
    not_blank = validate.Length(min=1, error="Field cannot be blank")

    id = fields.Integer(dump_only=True)
    news_id = fields.Integer(allow_none=True)
    sec_filing_id = fields.Integer(allow_none=True)
    company_filing_id = fields.Integer(allow_none=True)
    text_type = fields.String(allow_none=True)
    text = fields.String(required=True)
    preferred = fields.Boolean(allow_none=True)
    updated_at = fields.DateTime(dump_only=True)



from sqlalchemy.exc import SQLAlchemyError
from datetime import datetime
from marshmallow import ValidationError, EXCLUDE
from collections import defaultdict

from ...database import db_session
from ...resources.ClaimChallenged.models import (
    ClaimChallengedModel,
)
from .schemas import (
    ClaimChallengedResourceSchema,
    ClaimChallengedQueryParamsSchema,
)


schema_resource = ClaimChallengedResourceSchema()
schema_params = ClaimChallengedQueryParamsSchema()


class ClaimChallengedNotFoundException(Exception):
    def __init__(self, message):
        self.message = message


class ClaimChallenged:
    @staticmethod
    def create(params):
        """
        :param
            ClaimChallengedResourceSchema
        :return:
            ClaimChallengedResourceSchema
        :exception:
            ValidationError
        """
        data = schema_resource.load(params, unknown=EXCLUDE)
        if errors:
            raise ValidationError(errors)

        response = _helper_create(data)
        return response

    @staticmethod
    def read(params):
        """
        :param
            params: dict
                id
                claim_id
                prior_art_combination
                prior_art_id
        :return:
            queried object
        :exception:
            ValidationError
        """

        data = schema_params.load(params, unknown=EXCLUDE)

        claim_challenged_query = _build_query(params=data)
        response = schema_resource.dump(claim_challenged_query, many=True)
        return response

    @staticmethod
    def one(params):
        """
        :param
            params: dict
                id
                claim_id
                prior_art_combination
                prior_art_id
        :return:
            queried object
        :exception:
            ValidationError
        """

        data = schema_params.load(params, unknown=EXCLUDE)

        claim_challenged_query = _build_query(params=data).one()
        response = schema_resource.dump(claim_challenged_query)
        return response

    @staticmethod
    def delete(id):
        """
        :param
            id
        :return:
            delete message
        :exception:
            ClaimChallengedNotFoundException
        """

        claim_challenged_query = db_session.query(
            ClaimChallengedModel).filter_by(
            id=id).first()
        if not claim_challenged_query:
            raise ClaimChallengedNotFoundException(
                'Claim challenged does not exist!')
        try:
            db_session.delete(claim_challenged_query)
            db_session.commit()
            db_session.close()
            return 'Successfully deleted'
        except SQLAlchemyError:
            db_session.rollback()
            db_session.close()
            raise

    @staticmethod
    def bulk_delete(ptab2_proceeding_id):
        try:
            db_session.query(
                ClaimChallengedModel).filter(
                ClaimChallengedModel.ptab2_proceeding_id == ptab2_proceeding_id
            ).delete(synchronize_session=False)
            db_session.commit()
            db_session.close()
            return 'Successfully deleted'
        except SQLAlchemyError:
            db_session.rollback()
            db_session.close()
            raise

    def view_combinations(self, claim_id, ptab2_proceeding_id):
        params = {
            'claim_id': claim_id,
            'ptab2_proceeding_id': ptab2_proceeding_id,
        }
        ccs = self.read(params)
        result = defaultdict(list)
        for cc in ccs:
            result[cc['prior_art_combination']].append({
                'prior_art_id': cc['prior_art_id'],
                'nature': cc['prior_art_nature'],
            })
        return result


def _helper_create(data):
    new_claim_challenged = ClaimChallengedModel(
        ptab2_proceeding_id=data['ptab2_proceeding_id'],
        claim_id=data['claim_id'],
        prior_art_combination=data['prior_art_combination'],
        prior_art_id=data['prior_art_id'],
        prior_art_nature=data['prior_art_nature'],
        nature_of_challenge=data['nature_of_challenge'],
        updated_at=datetime.utcnow(),
    )
    try:
        db_session.add(new_claim_challenged)
        db_session.commit()
        claim_challenged_query = db_session.query(
            ClaimChallengedModel).get(
            new_claim_challenged.id)
        response = schema_resource.dump(claim_challenged_query)
        db_session.close()
        return response
    except SQLAlchemyError:
        db_session.rollback()
        db_session.close()
        raise


def _build_query(params):
    q = db_session.query(
        ClaimChallengedModel)
    if 'id' in params:
        q = q.filter_by(id=params['id'])
    if 'ptab2_proceeding_id' in params:
        q = q.filter_by(ptab2_proceeding_id=params['ptab2_proceeding_id'])
    if 'claim_id' in params:
        q = q.filter_by(claim_id=params['claim_id'])
    if 'prior_art_combination' in params:
        q = q.filter_by(
            prior_art_combination=params['prior_art_combination'])
    if 'prior_art_id' in params:
        q = q.filter_by(prior_art_id=params['prior_art_id'])
    if 'prior_art_nature' in params:
        q = q.filter_by(prior_art_nature=params['prior_art_nature'])
    return q



from marshmallow import (
    Schema,
    fields,
    validate,
)


class NCTFacilityResourceSchema(Schema):
    not_blank = validate.Length(min=1, error="Field cannot be blank")

    id = fields.Integer(dump_only=True)
    nct_study_id = fields.Integer(required=True)
    facility_name = fields.String(allow_none=True)
    facility_city = fields.String(allow_none=True)
    facility_country = fields.String(allow_none=True)
    facility_zip_code = fields.String(allow_none=True)
    updated_at = fields.DateTime(dump_only=True)



import copy

from marshmallow import ValidationError
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Query
from parameterized import parameterized

from test.backendtestcase import TestCase
from test.utils import second_equals_first
from src.cs_models.resources.ProceedingStage import (
    ProceedingStage,
)
from src.cs_models.resources.ProceedingStage.models import (
    ProceedingStageModel,
)
from src.cs_models.resources.Subsidiary import Subsidiary
from src.cs_models.resources.CompanySEC import CompanySEC
from src.cs_models.resources.CompanyOUS import CompanyOUS
from src.cs_models.resources.File import File
from src.cs_models.resources.PTAB2Proceeding import PTAB2Proceeding
from src.cs_models.database import db_session


class ProceedingStageResourceTestCase(TestCase):
    def setUp(self):
        super(ProceedingStageResourceTestCase, self).setUp()
        self.inst = ProceedingStage()
        self.inst_ptab2_proceeding = PTAB2Proceeding()
        self.inst_subsidiary = Subsidiary()
        self.inst_sec_company = CompanySEC()
        self.inst_ous_company = CompanyOUS()
        self.inst_file = File()

        self.file1 = self.inst_file.create(
            {
                "file_format": "pdf",
                "s3_bucket_name": "ptab-docs",
                "s3_key_name": "IPR2017-000894/scheduling_order.pdf",
            }
        )

        self.company_sec1 = self.inst_sec_company.create({
            'cik_str': '200406',
            'ticker': 'JNJ',
            "title": "JOHNSON & JOHNSON",
        })

        self.company_sec2 = self.inst_sec_company.create({
            "cik_str": "816284",
            "ticker": "CELG",
            "title": "CELGENE CORP /DE/"
        })

        self.company_ous1 = self.inst_ous_company.create({
            'name': 'Aurobindo Pharma',
            'ticker': 'AUROPHARMA',
            'exchange': 'NSE',
        })

        self.subsidiary1 = self.inst_subsidiary.create({
            'name': 'Janssen Scientific Affairs, LLC',
            'company_sec_id': self.company_sec1['id'],
            'company_ous_id': None,
        })

        self.subsidiary2 = self.inst_subsidiary.create({
            'name': 'Celgene International II SARL',
            'company_sec_id': self.company_sec2['id'],
            'company_ous_id': None,
        })
        self.ptab2_proceeding1 = self.inst_ptab2_proceeding.create({
            'accorded_filing_date': '05-06-2013',
            'decision_date': '10-07-2014',
            'institution_decision_date': '10-08-2013',
            'petitioner_application_number_text': '-',
            'petitioner_counsel_name': 'Steven Baughman',
            'petitioner_grant_date': '-',
            'petitioner_group_art_unit_number': '-',
            'petitioner_inventor_name': '-',
            'petitioner_party_name': 'Apple Inc.',
            'petitioner_patent_number': '-',
            'petitioner_patent_owner_name': '-',
            'petitioner_technology_center_number': '-',
            'proceeding_filing_date': '05-06-2013',
            'proceeding_last_modified_date': '-',
            'proceeding_number': 'IPR-100',
            'proceeding_status_category': 'FWD Entered',
            'proceeding_type_category': 'AIA Trial',
            'respondent_application_number_text': '08471964',
            'respondent_counsel_name': 'David Marsh',
            'respondent_grant_date': '10-12-1999',
            'respondent_group_art_unit_number': '-',
            'respondent_inventor_name': 'ARTHUR HAIR',
            'respondent_party_name': 'SightSound Technologies LLC',
            'respondent_patent_number': '5966440',
            'respondent_patent_owner_name': 'SightSound Technologies LLC',
            'respondent_technology_center_number': '2100',
            'subproceeding_type_category': 'CBM',
            'proceeding_status': 'FWD',
            'respondent_subsidiary_id': None,
            'petitioner_subsidiary_id': None,
        })

        self.proceeding_stage1 = self.inst.create({
            'ptab2_proceeding_id': self.ptab2_proceeding1['id'],
            'stage': 'Petition Filed',
            'is_active': True,
            'filed_date': '06-11-2019',
            'due_date': None,
        })

        self.valid_data = {
            'ptab2_proceeding_id': self.ptab2_proceeding1['id'],
            'stage': 'PO Preliminary Response',
            'is_active': True,
            'filed_date': '06-11-2019',
            'due_date': '06-20-2019',
        }

    @parameterized.expand([
        ('ptab2_proceeding_id',),
        ('stage',),
        ('is_active',),
    ])
    def test_create_validation_error_missing_field(self, field_to_pop):
        base_data = copy.copy(self.valid_data)
        base_data.pop(field_to_pop)
        self.assertRaises(
            ValidationError,
            self.inst.create,
            base_data,
        )

    def test_create_violates_unique_constraint(self):
        self.assertRaises(
            IntegrityError,
            self.inst.create,
            {
                'ptab2_proceeding_id': (
                    self.proceeding_stage1['ptab2_proceeding_id']),
                'stage': self.proceeding_stage1['stage'],
                'is_active': True,
            },
        )

    def test_create(self):
        resp = self.inst.create(self.valid_data)
        second_equals_first(
            self.valid_data,
            resp,
            fields_to_ignore=['filed_date', 'due_date'],
        )

    def test_delete(self):
        self.inst.delete(id=self.proceeding_stage1['id'])

        q = self._build_query(proceeding_stage_id=self.proceeding_stage1['id'])

        assert q.first() is None

    def test_bulk_delete(self):
        self.inst.create(self.valid_data)

        self.assertEqual(2, len(self._read_by_ptab2_proceeding_id(
            self.proceeding_stage1['ptab2_proceeding_id'])))

        self.inst.bulk_delete(
            ptab2_proceeding_id=self.proceeding_stage1['ptab2_proceeding_id'])

        self.assertEqual(0, len(self._read_by_ptab2_proceeding_id(
            self.proceeding_stage1['ptab2_proceeding_id'])))

    @staticmethod
    def _read_by_ptab2_proceeding_id(ptab2_proceeding_id):
        q = db_session.query(
            ProceedingStageModel,
        ).filter_by(ptab2_proceeding_id=ptab2_proceeding_id)
        return q.all()

    @staticmethod
    def _build_query(proceeding_stage_id: int) -> Query:
        q = db_session.query(
            ProceedingStageModel,
        ).filter_by(id=proceeding_stage_id)
        return q




from sqlalchemy import (
    Column,
    Integer,
    String,
    ForeignKey,
)

from ...aact_database import Base


class KeywordModel(Base):
    __tablename__ = 'keywords'
    __table_args__ = (
        {'schema': 'ctgov'},
    )
    id = Column(Integer, primary_key=True)
    nct_id = Column(
        String,
        ForeignKey('ctgov.studies.nct_id'),
        nullable=False,
    )
    name = Column(String)
    downcase_name = Column(String)

    def __repr__(self):
        return "<Keyword(id='{}', nct_id='{}', " \
               "name='{}', downcase_name='{}'>"\
            .format(self.id, self.nct_id, self.name, self.downcase_name)



from sqlalchemy.exc import SQLAlchemyError

from ...database import db_session
from ...resources.Claim.models import ClaimModel
from .schemas import (
    ClaimResourceSchema,
    ClaimQueryParamsSchema,
    ClaimPatchSchema,
)
from marshmallow import ValidationError, EXCLUDE


schema_resource = ClaimResourceSchema()
schema_params = ClaimQueryParamsSchema()
schema_patch = ClaimPatchSchema()


class DBException(SQLAlchemyError):
    pass


class ClaimNotFoundException(Exception):
    def __init__(self, message):
        self.message = message


class Claim:
    @staticmethod
    def create(params):
        """
        :param
            Dict
                patent_id: int
                claim_number: int
                claim_text: string
        :return:
            newly created object
        :exception:
            ValidationError
            DBException
        """

        data = schema_resource.load(params, unknown=EXCLUDE)
        response = _helper_create(data)
        return response

    @staticmethod
    def read(params):
        """
        :param
            params: dict
                id
                patent_id
        :return:
            queried object
        :exception:
            ValidationError
        """

        data = schema_params.load(params, unknown=EXCLUDE)

        claim_query = _build_query(params=data)
        response = schema_resource.dump(claim_query, many=True)
        return response

    @staticmethod
    def one(params):
        """
            :param
                params: dict
                    id
                    patent_id
            :return:
                queried object
            :exception:
                ValidationError
            """

        data = schema_params.load(params, unknown=EXCLUDE)

        claim_query = _build_query(params=data).one()
        response = schema_resource.dump(claim_query)
        return response

    @staticmethod
    def delete(id):
        """
        :param
            id
        :return:
            delete message
        :exception:
            ClaimNotFoundException
            DBException
        """

        claim_query = db_session.query(ClaimModel).filter_by(id=id).one()
        try:
            db_session.delete(claim_query)
            db_session.commit()
            db_session.close()
            return 'Successfully deleted'
        except SQLAlchemyError:
            db_session.rollback()
            db_session.close()
            raise


def _helper_create(data):
    new_claim = ClaimModel(**data)
    try:
        db_session.add(new_claim)
        db_session.commit()
        claim_query = db_session.query(ClaimModel).get(new_claim.id)
        response = schema_resource.dump(claim_query)
        db_session.close()
        return response
    except SQLAlchemyError:
        db_session.rollback()
        db_session.close()
        raise


def _build_query(params):
    q = db_session.query(ClaimModel)
    if params.get('id'):
        q = q.filter_by(id=params.get('id'))
    if params.get('patent_id'):
        q = q.filter_by(patent_id=params.get('patent_id'))
    if params.get('patent_application_id'):
        q = q.filter_by(
            patent_application_id=params.get('patent_application_id'))
    if params.get('claim_number'):
        q = q.filter_by(claim_number=params.get('claim_number'))
    return q



from sqlalchemy import (
    Column,
    Integer,
    String,
    Binary,
)

from ...database import Base


class DOILinkModel(Base):
    __tablename__ = "doi_links"

    id = Column(Integer, primary_key=True)
    link = Column(
        Binary(16),
        nullable=False,
        index=True,
    )
    doi = Column(String(191), nullable=False)
